import datetime
import pytest

from pyrego600.decoders import Decoders
from pyrego600.regoerror import RegoError


def test_expected_len():
    assert Decoders.ERROR.length == 42


def test_decode_empty_buffer():
    with pytest.raises(RegoError, match="Unexpected size of ''; expecting: 42, got: 0"):
        Decoders.ERROR.decode(b"")


def test_decode_non_empty_buffer_with_invalid_len():
    with pytest.raises(RegoError, match="Unexpected size of '00010203'; expecting: 42, got: 4"):
        Decoders.ERROR.decode(b"\x00\x01\x02\x03")


def test_decode_buffer_with_invalid_header():
    with pytest.raises(
        RegoError,
        match="Invalid header for '000000000000000000000000000000000000000000000000000000000000000000000000000000000000'",
    ):
        Decoders.ERROR.decode(bytes(42))


def test_decode_buffer_with_valid_header_and_invalid_checksum():
    with pytest.raises(
        RegoError,
        match="Invalid crc of '010200000000000000000000000000000000000000000000000000000000000000000000000000000000'",
    ):
        Decoders.ERROR.decode(b"\x01\x02" + bytes(40))


def test_decode_no_error():
    assert Decoders.ERROR.decode(b"\x01\xff" + bytes(39) + b"\xff") is None


def test_decode_value():
    last_error = Decoders.ERROR.decode(
        b"\x01\x00\x0b\x03\x02\x03\x05\x03\x00\x03\x01\x03\x02\x03\x08\x02\x00\x03\x00\x03\x05\x03\x0a\x03\x01\x03"
        + b"\x01\x03\x0a\x03\x05\x03\x02\x00\x00\x02\x01\x0b\x02\x01\x07\x0b",
    )
    assert last_error.code == 11
    assert last_error.timestamp == datetime.datetime(2025, 1, 28, 5, 11, 52)



from dataclasses import dataclass
from typing import NamedTuple

from .checksum import checksum
from .decoders import Decoder, Decoders
from .identifier import Identifier
from .source import Source
from .transformations import Transformation, Transformations
from .type import Type
from .value_converter import int16_to_seven_bit_format


@dataclass(frozen=True)
class Register:
    identifier: Identifier
    source: Source
    address: int
    decoder: Decoder
    transformation: Transformation
    type: Type
    is_writtable: bool = False

    class Command(NamedTuple):
        payload: bytes
        decoder: Decoder
        transformation: Transformation

    def _read(self) -> Command:
        return Register.Command(self.__payload(self.source.read, 0), self.decoder, self.transformation)

    def _write(self, data: int) -> Command:
        if self.source.write is None:
            raise TypeError
        return Register.Command(self.__payload(self.source.write, data), Decoders.EMPTY, Transformations.IDENTITY)

    def __payload(self, source: int, data: int) -> bytes:
        payload_bytes = int16_to_seven_bit_format(self.address) + int16_to_seven_bit_format(data)
        return b"\x81" + source.to_bytes() + payload_bytes + checksum(payload_bytes).to_bytes()



from dataclasses import dataclass

from .group import Group


@dataclass(frozen=True)
class Identifier:
    id: str
    group: Group

    def __str__(self):
        return f"{self.group.value}-{self.id}"



from abc import ABC

from ..last_error import LastError


class Decoder(ABC):
    @property
    def length(self) -> int:
        raise NotImplementedError

    def decode(self, buffer: bytes) -> int | LastError | None:
        raise NotImplementedError



import pytest

from pyrego600.decoders import Decoders
from pyrego600.regoerror import RegoError


def test_expected_len():
    assert Decoders.INT_16.length == 5


def test_decode_empty_buffer():
    with pytest.raises(RegoError, match="Unexpected size of ''; expecting: 5, got: 0"):
        Decoders.INT_16.decode(b"")


def test_decode_non_empty_buffer_with_invalid_len():
    with pytest.raises(RegoError, match="Unexpected size of '00010203'; expecting: 5, got: 4"):
        Decoders.INT_16.decode(b"\x00\x01\x02\x03")


def test_decode_buffer_with_invalid_header():
    with pytest.raises(RegoError, match="Invalid header for '0203040506'"):
        Decoders.INT_16.decode(b"\x02\x03\x04\x05\x06")


def test_decode_buffer_with_valid_header_and_invalid_checksum():
    with pytest.raises(RegoError, match="Invalid crc of '0103040506'"):
        Decoders.INT_16.decode(b"\x01\x03\x04\x05\x06")


def test_decode_positive_value():
    assert Decoders.INT_16.decode(b"\x01\x00\x02\x2c\x2e") == 300


def test_decode_negative_value():
    assert Decoders.INT_16.decode(b"\x01\x03\x7f\x77\x0b") == -9
    assert Decoders.INT_16.decode(b"\x01\x03\x7c\x1d\x62") == -483


def test_decodings():
    readings = [
        (b"\x01\x03\x7f\x77\x0b", -9),
        (b"\x01\x00\x02\x41\x43", 321),
        (b"\x01\x00\x03\x06\x05", 390),
        (b"\x01\x00\x02\x29\x2b", 297),
        (b"\x01\x00\x00\x20\x20", 32),
        (b"\x01\x00\x00\x06\x06", 6),
        (b"\x01\x00\x03\x0e\x0d", 398),
        (b"\x01\x00\x00\x00\x00", 0),
        (b"\x01\x00\x00\x01\x01", 1),
        (b"\x01\x00\x00\x1b\x1b", 27),
        (b"\x01\x00\x00\x64\x64", 100),
        (b"\x01\x01\x5c\x0a\x57", 28170),
        (b"\x01\x00\x20\x20\x00", 4128),
        (b"\x01\x00\x01\x22\x23", 162),
        (b"\x01\x00\x00\x3e\x3e", 62),
    ]
    for reading in readings:
        assert Decoders.INT_16.decode(reading[0]) == reading[1]



import datetime
from ..last_error import LastError
from ..value_converter import array_to_byte, string_from_bytes
from .abstract_decoder import AbstractDecoder


class ErrorDecoder(AbstractDecoder):
    @property
    def length(self) -> int:
        return 42

    def _convert(self, buffer: bytes) -> LastError | None:
        if buffer[1] == 0xFF:
            return None
        timestamp_str = string_from_bytes(buffer, 3, 15)
        year = int(timestamp_str[:2]) + 1000
        if year < 1950:
            year += 1000
        timestamp = datetime.datetime(
            year,
            int(timestamp_str[2:4]),
            int(timestamp_str[4:6]),
            int(timestamp_str[7:9]),
            int(timestamp_str[10:12]),
            int(timestamp_str[13:15]),
        )
        return LastError(array_to_byte(buffer, 1), timestamp)



from .empty_decoder import EmptyDecoder
from .error_decoder import ErrorDecoder
from .int16_decoder import Int16Decoder


class Decoders:
    EMPTY = EmptyDecoder()
    ERROR = ErrorDecoder()
    INT_16 = Int16Decoder()



from .group import Group
from .heatpump import HeatPump
from .identifier import Identifier
from .last_error import LastError
from .register import Register
from .regoerror import RegoError
from .type import Type

__all__ = ["Group", "HeatPump", "Identifier", "LastError", "Register", "RegoError", "Type"]



"""Provides TODO."""

import asyncio
import logging

from asyncio import timeout as asyncio_timeout
from typing import Self

from .connection import Connection
from .decoders import Decoder
from .register import Register
from .register_repository import RegisterRepository
from .regoerror import RegoError
from .serial_connection import SerialConnection
from .transformations import Transformation

_LOGGER = logging.getLogger(__name__)
_RETRIES: int = 3


class HeatPump:
    def __init__(self, connection: Connection) -> None:
        self.__connection = connection
        self.__lock = asyncio.Lock()

    @classmethod
    def connect(cls, url: str) -> Self:
        connection = SerialConnection(url)
        return cls(connection)

    @property
    def registers(self) -> list[Register]:
        """Return the register database."""
        return RegisterRepository.registers()

    async def dispose(self):
        await self.__connection.close()

    async def verify(self, retry: int = _RETRIES) -> None:
        _LOGGER.debug("Reading Rego device version...")
        register = RegisterRepository.version()
        async with self.__lock:
            version = await self.__send(*register._read(), retry)
            if version != 600:
                await self.__connection.close()
                raise RegoError(f"Invalid rego version received {version}")
        _LOGGER.debug("Connected to Rego version %s.", version)

    async def read(self, register: Register, retry: int = _RETRIES) -> float:
        async with self.__lock:
            return await self.__send(*register._read(), retry)

    async def write(self, register: Register, value: float, retry: int = _RETRIES) -> None:
        transformed = register.transformation.from_value(value)
        async with self.__lock:
            return await self.__send(*register._write(transformed), retry)

    async def __send(self, payload: bytes, decoder: Decoder, transformation: Transformation, retry: int) -> float:
        try:
            if not self.__connection.is_connected:
                _LOGGER.debug("Not connected, connecting...")
                await self.__connection.connect()
                _LOGGER.debug("Connected")

            # Give heat pump some time between commands. Feeding commands too quickly
            # might cause heat pump not to respond.
            await asyncio.sleep(0.05)

            async with asyncio_timeout(2):
                _LOGGER.debug("Sending '%s'", payload.hex())
                await self.__connection.write(payload)
                _LOGGER.debug("Send, waiting for response...")
                response = await self.__connection.read(decoder.length)
                _LOGGER.debug("Received %s", response.hex())
            return transformation.to_value(decoder.decode(response))

        except (OSError, RegoError) as e:
            _LOGGER.debug("Sending '%s' failed due %s", payload.hex(), repr({e}))
            await self.__connection.close()
            if retry > 0:
                _LOGGER.debug("Retrying, retry=%d", retry)
                await asyncio.sleep(0.2)
                return await self.__send(payload, decoder, transformation, retry - 1)
            raise



from abc import ABC


class Connection(ABC):
    @property
    def is_connected(self) -> bool:
        raise NotImplementedError

    async def connect(self) -> None:
        raise NotImplementedError

    async def close(self) -> None:
        raise NotImplementedError

    async def read(self, length: int) -> bytes:
        raise NotImplementedError

    async def write(self, buffer: bytes) -> None:
        raise NotImplementedError



from .decoders import Decoders
from .identifier import Identifier
from .register import Register
from .sources import Sources
from .transformations import Transformations
from .type import Type


class RegisterFactory:
    @staticmethod
    def version(identifier: Identifier) -> Register:
        return Register(
            identifier=identifier,
            source=Sources.VERSION,
            address=0x0000,
            decoder=Decoders.INT_16,
            transformation=Transformations.IDENTITY,
            type=None,
        )

    @staticmethod
    def last_error(identifier: Identifier) -> Register:
        return Register(
            identifier=identifier,
            source=Sources.LAST_ERROR,
            address=0x0000,
            decoder=Decoders.ERROR,
            transformation=Transformations.IDENTITY,
            type=Type.ERROR,
        )

    @staticmethod
    def front_panel_switch(identifier: Identifier, address: int) -> Register:
        return Register(
            identifier=identifier,
            source=Sources.FRONT_PANEL,
            address=address,
            decoder=Decoders.INT_16,
            transformation=Transformations.IDENTITY,
            type=Type.SWITCH,
        )

    @staticmethod
    def system_temperature(identifier: Identifier, address: int, is_writtable: bool = False) -> Register:
        return Register(
            identifier=identifier,
            source=Sources.SYSTEM,
            address=address,
            decoder=Decoders.INT_16,
            transformation=Transformations.NUMERIC_ONE_TENTH,
            type=Type.TEMPERATURE,
            is_writtable=is_writtable,
        )

    @staticmethod
    def system_unitless(identifier: Identifier, address: int, is_writtable: bool = False) -> Register:
        return Register(
            identifier=identifier,
            source=Sources.SYSTEM,
            address=address,
            decoder=Decoders.INT_16,
            transformation=Transformations.NUMERIC_ONE_TENTH,
            type=Type.UNITLESS,
            is_writtable=is_writtable,
        )

    @staticmethod
    def system_switch(identifier: Identifier, address: int) -> Register:
        return Register(
            identifier=identifier,
            source=Sources.SYSTEM,
            address=address,
            decoder=Decoders.INT_16,
            transformation=Transformations.IDENTITY,
            type=Type.SWITCH,
        )

    @staticmethod
    def system_hours(identifier: Identifier, address: int) -> Register:
        return Register(
            identifier=identifier,
            source=Sources.SYSTEM,
            address=address,
            decoder=Decoders.INT_16,
            transformation=Transformations.IDENTITY,
            type=Type.HOURS,
        )



import pytest

from pyrego600 import Identifier, Register
from pyrego600.identifiers import Identifiers
from pyrego600.register_repository import RegisterRepository


def verify_read(
    identifer: Identifier,
    expectedPayload: bytes,
    responseBytes: bytes,
    expectedDecodedValue: float,
    expectedValue: float,
) -> Register:
    register = next(r for r in RegisterRepository.registers() if r.identifier == identifer)
    command = register._read()
    assert command.payload == expectedPayload
    assert command.decoder.decode(responseBytes) == expectedDecodedValue
    assert command.transformation.to_value(expectedDecodedValue) == expectedValue
    return register


def test_sensor_values_radiator_return():
    verify_read(
        identifer=Identifiers.SENSOR_VALUES_RADIATOR_RETURN,
        expectedPayload=b"\x81\x02\x00\x04\x09\x00\x00\x00\x0d",
        responseBytes=b"\x01\x00\x02\x25\x27",
        expectedDecodedValue=293,
        expectedValue=29.3,
    )


def test_sensor_values_radiator_forward():
    verify_read(
        identifer=Identifiers.SENSOR_VALUES_RADIATOR_FORWARD,
        expectedPayload=b"\x81\x02\x00\x04\x0c\x00\x00\x00\x08",
        responseBytes=b"\x01\x00\x02\x41\x43",
        expectedDecodedValue=321,
        expectedValue=32.1,
    )


def test_sensor_values_outdoor():
    verify_read(
        identifer=Identifiers.SENSOR_VALUES_OUTDOOR,
        expectedPayload=b"\x81\x02\x00\x04\x0a\x00\x00\x00\x0e",
        responseBytes=b"\x01\x03\x7f\x5a\x26",
        expectedDecodedValue=-38,
        expectedValue=-3.8,
    )


def test_device_values_compressor():
    verify_read(
        identifer=Identifiers.DEVICE_VALUES_COMPRESSOR,
        expectedPayload=b"\x81\x02\x00\x03\x7e\x00\x00\x00\x7d",
        responseBytes=b"\x01\x00\x00\x01\x01",
        expectedDecodedValue=1,
        expectedValue=1,
    )


def test_settings_heat_curve():
    register = verify_read(
        identifer=Identifiers.SETTINGS_HEAT_CURVE,
        expectedPayload=b"\x81\x02\x00\x00\x00\x00\x00\x00\x00",
        responseBytes=b"\x01\x00\x00\x1b\x1b",
        expectedDecodedValue=27,
        expectedValue=2.7,
    )
    assert register.transformation.from_value(2.7) == 27
    command = register._write(27)
    assert command.payload == b"\x81\x03\x00\x00\x00\x00\x00\x1b\x1b"
    assert command.decoder.decode(b"\x01") == 0
    assert command.transformation.to_value(0) == 0
    assert str(register.identifier) == "settings-heat_curve"
    assert register.is_writtable


def test_front_panel_power_lamp():
    register = verify_read(
        identifer=Identifiers.FRONT_PANEL_POWER_LAMP,
        expectedPayload=b"\x81\x00\x00\x00\x12\x00\x00\x00\x12",
        responseBytes=b"\x01\x00\x00\x01\x01",
        expectedDecodedValue=1,
        expectedValue=1,
    )
    assert not register.is_writtable
    with pytest.raises(TypeError):
        register._write(0)


def test_sensor_values_indoor_not_available():
    verify_read(
        identifer=Identifiers.SENSOR_VALUES_INDOOR,
        expectedPayload=b"\x81\x02\x00\x04\x0d\x00\x00\x00\x09",
        responseBytes=b"\x01\x03\x7c\x1d\x62",
        expectedDecodedValue=-483,
        expectedValue=None,
    )



from .identity_transformation import IdentityTransformation
from .numeric_transformation import NumericTransformation


class Transformations:
    IDENTITY = IdentityTransformation()
    NUMERIC_ONE_TENTH = NumericTransformation(multiplier=0.1)



from ..checksum import checksum
from ..last_error import LastError
from ..regoerror import RegoError
from .decoder import Decoder


class AbstractDecoder(Decoder):
    def _convert(self, buffer: bytes) -> int | LastError | None:
        raise NotImplementedError

    def decode(self, buffer: bytes) -> int | LastError | None:
        if self.length < 1 or len(buffer) != self.length:
            raise RegoError(f"Unexpected size of '{buffer.hex()}'; expecting: {self.length}, got: {len(buffer)}")

        if buffer[0] != 0x01:
            raise RegoError(f"Invalid header for '{buffer.hex()}'")

        if self.length > 1 and checksum(buffer[1 : self.length - 1]) != buffer[self.length - 1]:
            raise RegoError(f"Invalid crc of '{buffer.hex()}'")

        return self._convert(buffer)



import serial_asyncio_fast as serial_asyncio

from .connection import Connection


class SerialConnection(Connection):
    def __init__(self, url) -> None:
        self.__url = url
        self.__reader = None
        self.__writter = None

    @property
    def is_connected(self) -> bool:
        return self.__reader is not None

    async def connect(self) -> None:
        self.__reader, self.__writter = await serial_asyncio.open_serial_connection(url=self.__url, baudrate=19200)

    async def close(self) -> None:
        if self.__writter is not None:
            self.__writter.close()
            await self.__writter.wait_closed()
        self.__writter = None
        self.__reader = None

    async def read(self, length: int) -> bytes:
        return await self.__reader.readexactly(length)

    async def write(self, buffer: bytes) -> None:
        self.__writter.write(buffer)
        await self.__writter.drain()



from .abstract_decoder import AbstractDecoder


class EmptyDecoder(AbstractDecoder):
    @property
    def length(self) -> int:
        return 1

    def _convert(self, buffer: bytes) -> int:
        return 0



from abc import ABC

from ..last_error import LastError


class Transformation(ABC):
    def to_value(self, value: int | LastError | None) -> int | LastError | None:
        raise NotImplementedError

    def from_value(self, value: float) -> int:
        raise NotImplementedError



from .group import Group
from .identifier import Identifier


class Identifiers:
    # HP version
    VERSION = Identifier(id="version", group=Group.INFO)
    # Last error information
    LAST_ERROR = Identifier(id="last_error", group=Group.INFO)
    # Front Panel
    FRONT_PANEL_POWER_LAMP = Identifier(id="power_lamp", group=Group.FRONT_PANEL)
    FRONT_PANEL_PUMP_LAMP = Identifier(id="heat_pump_lamp", group=Group.FRONT_PANEL)
    FRONT_PANEL_ADDITIONAL_HEAT_LAMP = Identifier(id="additional_heat_lamp", group=Group.FRONT_PANEL)
    FRONT_PANEL_WATER_HEATER_LAMP = Identifier(id="hot_water_lamp", group=Group.FRONT_PANEL)
    FRONT_PANEL_ALARM_LAMP = Identifier(id="alarm_lamp", group=Group.FRONT_PANEL)
    # System registers, sensor values
    SENSOR_VALUES_RADIATOR_RETURN = Identifier(id="radiator_return", group=Group.SENSOR_VALUES)
    SENSOR_VALUES_OUTDOOR = Identifier(id="outdoor", group=Group.SENSOR_VALUES)
    SENSOR_VALUES_HOTWATER = Identifier(id="hot_water", group=Group.SENSOR_VALUES)
    SENSOR_VALUES_RADIATOR_FORWARD = Identifier(id="radiator_forward", group=Group.SENSOR_VALUES)
    SENSOR_VALUES_INDOOR = Identifier(id="indoor", group=Group.SENSOR_VALUES)
    SENSOR_VALUES_COMPRESSOR = Identifier(id="compressor", group=Group.SENSOR_VALUES)
    SENSOR_VALUES_HEATFLUID_OUT = Identifier(id="heat_fluid_out", group=Group.SENSOR_VALUES)
    SENSOR_VALUES_HEATFLUID_IN = Identifier(id="heat_fluid_in", group=Group.SENSOR_VALUES)
    SENSOR_VALUES_COLDFLUID_IN = Identifier(id="cold_fluid_in", group=Group.SENSOR_VALUES)
    SENSOR_VALUES_COLDFLUID_OUT = Identifier(id="cold_fluid_out", group=Group.SENSOR_VALUES)
    SENSOR_VALUES_EXTERNAL_HOTWATER = Identifier(id="external_hot_water", group=Group.SENSOR_VALUES)
    # System registers, control data
    CONTROL_DATA_RADIATOR_RETURN_TARGET = Identifier(id="radiator_return_target", group=Group.CONTROL_DATA)
    CONTROL_DATA_RADIATOR_RETURN_ON = Identifier(id="radiator_return_on", group=Group.CONTROL_DATA)
    CONTROL_DATA_RADIATOR_RETURN_OFF = Identifier(id="radiator_return_off", group=Group.CONTROL_DATA)
    CONTROL_DATA_HOTWATER_ON = Identifier(id="hot_water_on", group=Group.CONTROL_DATA)
    CONTROL_DATA_HOTWATER_OFF = Identifier(id="hot_water_off", group=Group.CONTROL_DATA)
    CONTROL_DATA_RADIATOR_FORWARD_TARGET = Identifier(id="radiator_forward_target", group=Group.CONTROL_DATA)
    CONTROL_DATA_ADDHEAT_POWER_PERCENT = Identifier(id="add_heat_power_percent", group=Group.CONTROL_DATA)
    # System registers, device values
    DEVICE_VALUES_COLD_FLUID_PUMP = Identifier(id="cold_fluid_pump", group=Group.DEVICE_VALUES)
    DEVICE_VALUES_COMPRESSOR = Identifier(id="compressor", group=Group.DEVICE_VALUES)
    DEVICE_VALUES_ADDITIONAL_HEAT_3KW = Identifier(id="additional_heat_3kw", group=Group.DEVICE_VALUES)
    DEVICE_VALUES_ADDITIONAL_HEAT_6KW = Identifier(id="additional_heat_6kw", group=Group.DEVICE_VALUES)
    DEVICE_VALUES_RADIATOR_PUMP = Identifier(id="radiator_pump", group=Group.DEVICE_VALUES)
    DEVICE_VALUES_HEATFLUID_PUMP = Identifier(id="heat_fluid_pump", group=Group.DEVICE_VALUES)
    DEVICE_VALUES_SWITCH_VALVE = Identifier(id="switch_valve", group=Group.DEVICE_VALUES)
    DEVICE_VALUES_ALARM = Identifier(id="alarm", group=Group.DEVICE_VALUES)
    # System registers, settings
    SETTINGS_HEAT_CURVE = Identifier(id="heat_curve", group=Group.SETTINGS)
    SETTINGS_HEAT_CURVE_FINE_ADJ = Identifier(id="heat_curve_fine_adj", group=Group.SETTINGS)
    SETTINGS_HEAT_CURVE_COUPLING_DIFF = Identifier(id="heat_curve_coupling_diff", group=Group.SETTINGS)
    SETTINGS_HEAT_CURVE_2 = Identifier(id="heat_curve_2", group=Group.SETTINGS)
    SETTINGS_HEAT_CURVE_2_FINE_ADJ = Identifier(id="heat_curve_2_fine_adj", group=Group.SETTINGS)
    SETTINGS_ADJ_CURVE_AT_20 = Identifier(id="adj_curve_at_20", group=Group.SETTINGS)
    SETTINGS_ADJ_CURVE_AT_15 = Identifier(id="adj_curve_at_15", group=Group.SETTINGS)
    SETTINGS_ADJ_CURVE_AT_10 = Identifier(id="adj_curve_at_10", group=Group.SETTINGS)
    SETTINGS_ADJ_CURVE_AT_5 = Identifier(id="adj_curve_at_5", group=Group.SETTINGS)
    SETTINGS_ADJ_CURVE_AT_0 = Identifier(id="adj_curve_at_0", group=Group.SETTINGS)
    SETTINGS_ADJ_CURVE_AT_MINUS_5 = Identifier(id="adj_curve_at_minus_5", group=Group.SETTINGS)
    SETTINGS_ADJ_CURVE_AT_MINUS_10 = Identifier(id="adj_curve_at_minus_10", group=Group.SETTINGS)
    SETTINGS_ADJ_CURVE_AT_MINUS_15 = Identifier(id="adj_curve_at_minus_15", group=Group.SETTINGS)
    SETTINGS_ADJ_CURVE_AT_MINUS_20 = Identifier(id="adj_curve_at_minus_20", group=Group.SETTINGS)
    SETTINGS_ADJ_CURVE_AT_MINUS_25 = Identifier(id="adj_curve_at_minus_25", group=Group.SETTINGS)
    SETTINGS_ADJ_CURVE_AT_MINUS_30 = Identifier(id="adj_curve_at_minus_30", group=Group.SETTINGS)
    SETTINGS_ADJ_CURVE_AT_MINUS_35 = Identifier(id="adj_curve_at_minus_35", group=Group.SETTINGS)
    SETTINGS_INDOOR_TEMP_SETTING = Identifier(id="indoor_temp_setting", group=Group.SETTINGS)
    SETTINGS_CURVE_INFL_BY_IN_TEMP = Identifier(id="curve_infl_by_in_temp", group=Group.SETTINGS)
    SETTINGS_SUMMER_DISCONNECTION = Identifier(id="summer_disconnection", group=Group.SETTINGS)
    SETTINGS_HOTWATER_TARGET = Identifier(id="hot_water_target", group=Group.SETTINGS)
    SETTINGS_HOTWATER_TARGET_HYSTERESIS = Identifier(id="hot_water_target_hysteresis", group=Group.SETTINGS)
    # System registers, operating times
    OPERATING_TIMES_HP_IN_OPERATION_RAD = Identifier(id="heat_pump_in_operation_rad", group=Group.OPERATING_TIMES)
    OPERATING_TIMES_HP_IN_OPERATION_DHW = Identifier(id="heat_pump_in_operation_dhw", group=Group.OPERATING_TIMES)
    OPERATING_TIMES_ADD_HEAT_IN_OPERATION_RAD = Identifier(id="add_heat_in_operation_rad", group=Group.OPERATING_TIMES)
    OPERATING_TIMES_ADD_HEAT_IN_OPERATION_DHW = Identifier(id="add_heat_in_operation_dhw", group=Group.OPERATING_TIMES)



import asyncio
from unittest import mock
from unittest.mock import PropertyMock, call

import pytest

from pyrego600 import HeatPump, RegoError
from pyrego600.connection import Connection
from pyrego600.identifiers import Identifiers
from pyrego600.register_factory import RegisterFactory

_REGISTER = RegisterFactory.system_temperature(Identifiers.SENSOR_VALUES_INDOOR, 0x020D)


@pytest.mark.asyncio
async def test_connect_fails():
    connection = mock.create_autospec(Connection)
    type(connection).is_connected = PropertyMock(return_value=False)
    connection.connect.side_effect = OSError
    hp = HeatPump(connection)
    with pytest.raises(OSError):
        await hp.read(_REGISTER, retry=0)
    connection.close.assert_called_once()


@pytest.mark.asyncio
@pytest.mark.parametrize("readReturnValue", [b"\xca\xfe\xba\xbe", b"\x01\x00\x00\x00\x00", b""])
async def test_verify_fails(readReturnValue):
    connection = mock.create_autospec(Connection)
    type(connection).is_connected = PropertyMock(return_value=False)
    connection.read.return_value = readReturnValue
    hp = HeatPump(connection)
    with pytest.raises(RegoError):
        await hp.verify(retry=0)
    connection.connect.assert_called_once()
    connection.close.assert_called_once()


@pytest.mark.asyncio
async def test_verify():
    connection = mock.create_autospec(Connection)
    type(connection).is_connected = PropertyMock(return_value=False)
    connection.read.return_value = b"\x01\x00\x04\x58\x5c"
    hp = HeatPump(connection)
    await hp.verify()
    connection.write.assert_called_once_with(b"\x81\x7f\x00\x00\x00\x00\x00\x00\x00")
    connection.connect.assert_called_once()
    connection.close.assert_not_called()
    await hp.dispose()
    connection.close.assert_called_once()


@pytest.mark.asyncio
async def test_waiting_for_response_should_timeout():
    async def looongRead(length: int):
        await asyncio.sleep(5)

    connection = mock.create_autospec(Connection)
    type(connection).is_connected = PropertyMock(return_value=False)
    connection.read.side_effect = looongRead
    hp = HeatPump(connection)
    with pytest.raises(TimeoutError):
        await hp.read(_REGISTER, retry=0)
    connection.connect.assert_called_once()
    connection.close.assert_called_once()


@pytest.mark.asyncio
async def test_waiting_for_response_should_timeout_and_retry():
    connection = mock.create_autospec(Connection)

    async def read(length: int):
        if len(connection.read.mock_calls) > 1:
            return b"\x01\x00\x00\x09\x09"
        await asyncio.sleep(5)

    type(connection).is_connected = PropertyMock(return_value=False)
    connection.read.side_effect = read
    hp = HeatPump(connection)
    assert await hp.read(_REGISTER, retry=1) == 0.9
    assert len(connection.connect.mock_calls) == 2
    connection.close.assert_called_once()


@pytest.mark.asyncio
async def test_read():
    connection = mock.create_autospec(Connection)
    type(connection).is_connected = PropertyMock(side_effect=lambda: len(connection.connect.mock_calls) > 0)
    connection.read.side_effect = [
        b"\x01\x00\x00\x08\x08",
        b"\x01\x00\x00\x0a\x0a",
    ]
    hp = HeatPump(connection)
    assert await hp.read(_REGISTER) == 0.8
    assert await hp.read(_REGISTER) == 1
    connection.connect.assert_called_once()
    connection.write.assert_has_calls(
        [
            call(b"\x81\x02\x00\x04\x0d\x00\x00\x00\x09"),
            call(b"\x81\x02\x00\x04\x0d\x00\x00\x00\x09"),
        ],
    )
    connection.close.assert_not_called()
    await hp.dispose()
    connection.close.assert_called_once()


@pytest.mark.asyncio
async def test_write():
    connection = mock.create_autospec(Connection)
    type(connection).is_connected = PropertyMock(return_value=True)
    connection.read.return_value = b"\x01"
    hp = HeatPump(connection)
    await hp.write(_REGISTER, 1.2)
    connection.connect.assert_not_called()
    connection.write.assert_called_once_with(b"\x81\x03\x00\x04\x0d\x00\x00\x0c\x05")
    connection.close.assert_not_called()
    await hp.dispose()
    connection.close.assert_called_once()



from dataclasses import dataclass

from .transformation import Transformation


@dataclass(frozen=True)
class NumericTransformation(Transformation):
    multiplier: float

    def to_value(self, value: int) -> float:
        # This value marks "absence" of a sensor
        if value == -483:
            return None
        return round(value * self.multiplier * 1 / self.multiplier) / (1 / self.multiplier)

    def from_value(self, value: float) -> int:
        return round(value / self.multiplier)



from .identifiers import Identifiers
from .register import Register
from .register_factory import RegisterFactory


class RegisterRepository:
    @staticmethod
    def version() -> Register:
        return RegisterFactory.version(identifier=Identifiers.VERSION)

    @staticmethod
    def registers() -> list[Register]:
        return [
            # Last error
            RegisterFactory.last_error(identifier=Identifiers.LAST_ERROR),
            # Front Panel
            RegisterFactory.front_panel_switch(
                identifier=Identifiers.FRONT_PANEL_POWER_LAMP,
                address=0x0012,
            ),
            RegisterFactory.front_panel_switch(
                identifier=Identifiers.FRONT_PANEL_PUMP_LAMP,
                address=0x0013,
            ),
            RegisterFactory.front_panel_switch(
                identifier=Identifiers.FRONT_PANEL_ADDITIONAL_HEAT_LAMP,
                address=0x0014,
            ),
            RegisterFactory.front_panel_switch(
                identifier=Identifiers.FRONT_PANEL_WATER_HEATER_LAMP,
                address=0x0015,
            ),
            RegisterFactory.front_panel_switch(
                identifier=Identifiers.FRONT_PANEL_ALARM_LAMP,
                address=0x0016,
            ),
            # System registers, sensor values
            RegisterFactory.system_temperature(
                identifier=Identifiers.SENSOR_VALUES_RADIATOR_RETURN,
                address=0x0209,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SENSOR_VALUES_OUTDOOR,
                address=0x020A,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SENSOR_VALUES_HOTWATER,
                address=0x020B,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SENSOR_VALUES_RADIATOR_FORWARD,
                address=0x020C,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SENSOR_VALUES_INDOOR,
                address=0x020D,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SENSOR_VALUES_COMPRESSOR,
                address=0x020E,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SENSOR_VALUES_HEATFLUID_OUT,
                address=0x020F,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SENSOR_VALUES_HEATFLUID_IN,
                address=0x0210,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SENSOR_VALUES_COLDFLUID_IN,
                address=0x0211,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SENSOR_VALUES_COLDFLUID_OUT,
                address=0x0212,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SENSOR_VALUES_EXTERNAL_HOTWATER,
                address=0x0213,
            ),
            # System registers, control data
            # TODO: check transformation and unit?
            # RegisterFactory.system_temperature(
            #    identifier=Identifiers.CONTROL_DATA_ADDHEAT_POWER_PERCENT,
            #    address=0x006C,
            # ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.CONTROL_DATA_RADIATOR_FORWARD_TARGET,
                address=0x006D,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.CONTROL_DATA_RADIATOR_RETURN_TARGET,
                address=0x006E,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.CONTROL_DATA_RADIATOR_RETURN_ON,
                address=0x006F,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.CONTROL_DATA_RADIATOR_RETURN_OFF,
                address=0x0070,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.CONTROL_DATA_HOTWATER_ON,
                address=0x0073,
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.CONTROL_DATA_HOTWATER_OFF,
                address=0x0074,
            ),
            # System registers, device values
            RegisterFactory.system_switch(
                identifier=Identifiers.DEVICE_VALUES_COLD_FLUID_PUMP,
                address=0x01FD,
            ),
            RegisterFactory.system_switch(
                identifier=Identifiers.DEVICE_VALUES_COMPRESSOR,
                address=0x01FE,
            ),
            RegisterFactory.system_switch(identifier=Identifiers.DEVICE_VALUES_ADDITIONAL_HEAT_3KW, address=0x01FF),
            RegisterFactory.system_switch(
                identifier=Identifiers.DEVICE_VALUES_ADDITIONAL_HEAT_6KW,
                address=0x0200,
            ),
            RegisterFactory.system_switch(
                identifier=Identifiers.DEVICE_VALUES_RADIATOR_PUMP,
                address=0x0203,
            ),
            RegisterFactory.system_switch(
                identifier=Identifiers.DEVICE_VALUES_HEATFLUID_PUMP,
                address=0x0204,
            ),
            RegisterFactory.system_switch(
                identifier=Identifiers.DEVICE_VALUES_SWITCH_VALVE,
                address=0x0205,
            ),
            RegisterFactory.system_switch(
                identifier=Identifiers.DEVICE_VALUES_ALARM,
                address=0x0206,
            ),
            # System registers, settings
            RegisterFactory.system_unitless(
                identifier=Identifiers.SETTINGS_HEAT_CURVE, address=0x0000, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_HEAT_CURVE_FINE_ADJ, address=0x0001, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_HEAT_CURVE_COUPLING_DIFF, address=0x0002, is_writtable=True
            ),
            RegisterFactory.system_unitless(
                identifier=Identifiers.SETTINGS_HEAT_CURVE_2, address=0x0003, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_HEAT_CURVE_2_FINE_ADJ, address=0x0004, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_ADJ_CURVE_AT_20, address=0x001E, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_ADJ_CURVE_AT_15, address=0x001C, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_ADJ_CURVE_AT_10, address=0x001A, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_ADJ_CURVE_AT_5, address=0x0018, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_ADJ_CURVE_AT_0, address=0x0016, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_ADJ_CURVE_AT_MINUS_5, address=0x0014, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_ADJ_CURVE_AT_MINUS_10, address=0x0012, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_ADJ_CURVE_AT_MINUS_15, address=0x0010, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_ADJ_CURVE_AT_MINUS_20, address=0x000E, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_ADJ_CURVE_AT_MINUS_25, address=0x000C, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_ADJ_CURVE_AT_MINUS_30, address=0x000A, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_ADJ_CURVE_AT_MINUS_35, address=0x0008, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_INDOOR_TEMP_SETTING, address=0x0021, is_writtable=True
            ),
            RegisterFactory.system_unitless(
                identifier=Identifiers.SETTINGS_CURVE_INFL_BY_IN_TEMP, address=0x0022, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_SUMMER_DISCONNECTION, address=0x0024, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_HOTWATER_TARGET, address=0x002B, is_writtable=True
            ),
            RegisterFactory.system_temperature(
                identifier=Identifiers.SETTINGS_HOTWATER_TARGET_HYSTERESIS, address=0x002C, is_writtable=True
            ),
            # System registers, operating times
            RegisterFactory.system_hours(identifier=Identifiers.OPERATING_TIMES_HP_IN_OPERATION_RAD, address=0x0048),
            RegisterFactory.system_hours(identifier=Identifiers.OPERATING_TIMES_HP_IN_OPERATION_DHW, address=0x004A),
            RegisterFactory.system_hours(
                identifier=Identifiers.OPERATING_TIMES_ADD_HEAT_IN_OPERATION_RAD,
                address=0x004C,
            ),
            RegisterFactory.system_hours(
                identifier=Identifiers.OPERATING_TIMES_ADD_HEAT_IN_OPERATION_DHW,
                address=0x004E,
            ),
        ]



def int16_to_seven_bit_format(value: int) -> bytes:
    return bytes([(value & 0xC000) >> 14, (value & 0x3F80) >> 7, value & 0x007F])


def seven_bit_format_to_int16(buffer: bytes, offset: int) -> int:
    value = buffer[offset] << 14 | buffer[offset + 1] << 7 | buffer[offset + 2]
    return value if value & 0x8000 == 0 else -(1 + (~value & 0x7FFF))


def array_to_byte(buffer: bytes, offset: int) -> int:
    return (buffer[offset] << 4 | buffer[offset + 1]) & 0xFF


def string_from_bytes(buffer: bytes, offset: int, charCount: int) -> str:
    return "".join(
        map(
            lambda i: chr(array_to_byte(buffer, i)),
            range(offset, offset + charCount * 2, 2),
        ),
    )



from ..value_converter import seven_bit_format_to_int16
from .abstract_decoder import AbstractDecoder


class Int16Decoder(AbstractDecoder):
    @property
    def length(self) -> int:
        return 5

    def _convert(self, buffer: bytes) -> int:
        return seven_bit_format_to_int16(buffer, 1)



from enum import Enum


class Group(Enum):
    SENSOR_VALUES = "sensor_values"
    CONTROL_DATA = "control_data"
    DEVICE_VALUES = "device_values"
    SETTINGS = "settings"
    OPERATING_TIMES = "operating_times"
    INFO = "info"
    FRONT_PANEL = "front_panel"



from ..last_error import LastError
from .transformation import Transformation


class IdentityTransformation(Transformation):
    def to_value(self, value: int | LastError | None) -> int | LastError | None:
        return value

    def from_value(self, value: float) -> int:
        return value



import asyncio
import logging

from pyrego600 import HeatPump, SerialConnection


async def hp_example():
    logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG)

    connection = SerialConnection(url="socket://192.168.2.50:9265")
    hp = HeatPump(connection)

    try:
        print("Connecting to Heat Pump...")
        await hp.verify()
        print("Connected!")

        for register in hp.registers:
            value = await hp.read(register)
            print(f"register {register.identifier} = {value}")
    except Exception as e:
        print(f"Reading registries from Heat Pump failed due {e!r}")
    finally:
        await hp.dispose()


if __name__ == "__main__":
    asyncio.set_event_loop(asyncio.new_event_loop())
    asyncio.run(hp_example())
